@testset "reported issues" begin
    @testset "barplot nl in text (#102)" begin
        p = barplot(
            ["I need a\nbreak", "I don't but could", :Hello, :Again],
            [30, 40, 20, 10],
        )
        test_ref("issues/barplot_nl.txt", @show_col(p))
    end

    @testset "plot hang (tan → ∞) (#129)" begin
        p = lineplot([cos, sin, tan], -π / 2, 2π)
        test_ref("issues/cos_sin_tan.txt", @show_col(p))
    end

    @testset "integer edges (#139)" begin
        p_int = horizontal_histogram(fit(Histogram, rand(10), 0:3))
        p_float = horizontal_histogram(fit(Histogram, rand(10), 0.0:1.0:3.0))
        @test @print_col(p_int) == @print_col(p_float)
    end

    @testset "isolated colorbar (#169)" begin
        p = Plot([0], [0], colorbar = true, colormap = :cividis, width = 0)
        test_ref("issues/isolated_colorbar.txt", @show_col(p))
    end

    @testset "steps (#211)" begin
        x = 1:5
        y = [1.0, 2.0, 3.0, 2.0, 1.0]
        p = scatterplot(x, y, marker = :circle)
        x = [1.0, 1.0, 2.0, 2.0, 3.0, 3.0, 4.0, 4.0, 5.0]
        y = [1.0, 2.0, 2.0, 3.0, 3.0, 2.0, 2.0, 1.0, 1.0]
        p = lineplot!(p, x, y)
        test_ref("issues/steps.txt", @show_col(p))
    end

    @testset "signed zero (#229)" begin
        p = histogram([-1.0, -0.0, 1.0])
        test_ref("issues/signed_zero.txt", @show_col(p))
    end

    @testset "grid log scale (#265)" begin
        @test lineplot(0.01:10, 0.01:10, xscale = :log10, yscale = :log10) isa Plot
    end

    @testset "Float input (#277)" begin
        for FX ∈ (Float16, Float32, Float64), FY ∈ (Float16, Float32, Float64)
            @test lineplot(FX[-1, 2], FY[-1, 9]) isa Plot
        end
        @test barplot(["Paris", "Madrid"], Float32[2.2, 8.4]; maximum = 10.0f0) isa Plot
    end
end
